// GivePrivIoXP.c

#include <stdio.h>
#include <string.h>
#include <windows.h>
#include <winioctl.h>
#include <process.h>

#include "PrivIoXP_IOCTL.h"
#include "SetPrivIoXP.h"

#define EGO "GivePrivIoXP"   /* our name */

static BOOL bVerbose = FALSE;

static void usage(void)
{
    fprintf(stderr, "Grant progam direct I/O port access (Windows NT/2000/XP)\n");
    fprintf(stderr, "usage: "EGO" [<addresses>] [<switches>] <executable> [<parameters>]\n");
    fprintf(stderr, "switches: -a : grant access to ALL I/O ports\n");
    fprintf(stderr, "          -v : verbose logging\n");
    fprintf(stderr, "Note: All <addresses> must begin with \"0x\"\n");
    exit(1);
}

int main(int argc, char **argv)
{
    char *arg;
    HANDLE hDriver = NULL;
    int tries;
    int count;
    DWORD value;
    char szRunString[256] = { "" };
    BOOL bExe;
    BOOL bAnyPrivs;
    DWORD nBytesRet;
    STARTUPINFO si;
    PROCESS_INFORMATION pi;

    setvbuf(stderr, (char *)NULL, _IONBF, 0); // unbuffer output

    if (argc <= 1) usage();

    if (!bIsWindowsNT()) {
        fprintf(stderr, EGO": Only works with Windows NT (Windows NT/2000/XP)!\n");
        return 1;
    }

    if (bVerbose) fprintf(stderr, EGO": using "DRIVER_NAME" driver\n");

    // open driver
    if (!(hDriver = OpenPrivIoXP(bVerbose, 
                               EGO": ", TRUE, 0, NULL, NULL, NULL))) {
        goto abort;
    }

    // default to restrict access to all ports
    if (!DeviceIoControl(hDriver,
			    IOCTL_IOPM_RESTRICT_ALL_ACCESS,
			    NULL, 0, NULL, 0, &nBytesRet, NULL)) {
        ShowErr(EGO": DeviceIoControl(IOCTL_IOPM_RESTRICT_ALL_ACCESS)");
	fprintf(stderr, EGO": error on IOCTL_IOPM_RESTRICT_ALL_ACCESS\n");
    	goto abort;
    }

    // process arguments and send to driver
    bExe = FALSE;
    bAnyPrivs = FALSE;
    for (count = 1; count < argc; count++) {
        arg = argv[count];
	// argument starts with '0x'?
	if (!bExe && *arg == '0' && arg[1] == 'x') {
	    sscanf(arg, "%x", &value);
	    if (!DeviceIoControl(hDriver,
				    IOCTL_SET_IOPM,
				    &value, sizeof(value),
                                    NULL, 0, &nBytesRet, NULL)) {
                ShowErr(EGO": DeviceIoControl(IOCTL_SET_IOPM)");
		fprintf(stderr, 
                    EGO": error granting direct access to I/O address 0x%03x\n",
		    value);
    	        goto abort;
	    } else {
		if (bVerbose) fprintf(stderr, 
                    EGO": I/O address 0x%03x granted direct access\n",
		    value);
                bAnyPrivs = TRUE;
            }
	} else if (!bExe && (!strcmp(arg, "/a") || !strcmp(arg, "-a"))) {
	    if (bVerbose) fprintf(stderr,
                          EGO": granting direct access to all I/O ports\n");
	    if (!DeviceIoControl(hDriver,
				    IOCTL_IOPM_ALLOW_ALL_ACCESS,
				    NULL,
				    0, NULL, 0, &nBytesRet, NULL)) {
                ShowErr(EGO": DeviceIoControl(IOCTL_IOPM_ALLOW_EXCUSIVE_ACCESS)");
		if (bVerbose) fprintf(stderr, 
                    EGO": error granting direct access to all I/O addresses\n");
    	        goto abort;
            }
            bAnyPrivs = TRUE;
	} else if (!bExe && (!strcmp(arg, "/v") || !strcmp(arg, "-v"))) {
            bVerbose = TRUE;
	} else if (!bExe && (!strcmp(arg, "/?") || !strcmp(arg, "-?") ||
	                     !strcmp(arg, "/h") || !strcmp(arg, "-h") ||
	                     !strcmp(arg, "-help"))) {
            usage();
	} else {
	    if (strlen(szRunString) + strlen(arg) + 4 >= sizeof(szRunString)) {
		fprintf(stderr, EGO": command line exceeds %d characters!\n",
                                sizeof(szRunString));
		goto abort;
	    }
            if (bExe) {
	        strcat(szRunString, " ");
            }
	    strcat(szRunString, "\"");
	    strcat(szRunString, arg);
	    strcat(szRunString, "\"");
            bExe = TRUE;
	}
    }

    if (!bExe) {
	fprintf(stderr, EGO": missing executable name!\n");
        goto abort;
    }

    if (!bAnyPrivs) {
	fprintf(stderr, EGO": Warning! no direct I/O access rights requested!\n");
    }

    // run user's program
    fprintf(stderr, "%s\n", szRunString);
    ZeroMemory(&si, sizeof(si));
    si.cb = sizeof(si);
    if (!CreateProcess(NULL,	// No module name (use command line).
		       szRunString,// Command line.
		       NULL,	// Process handle not inheritable.
		       NULL,	// Thread handle not inheritable.
		       FALSE,	// Set handle inheritance to FALSE.
		       0,	// No creation flags.
		       NULL,	// Use parent's environment block.
		       NULL,	// Use parent's starting directory.
		       &si,	// Pointer to STARTUPINFO structure.
		       &pi)	// Pointer to PROCESS_INFORMATION structure.
	) {
        ShowErr(EGO": CreateProcess");
	fprintf(stderr, EGO": CreateProcess error! (%s)\n", szRunString);
        goto abort;
    } else {
        if (bVerbose) fprintf(stderr, EGO": process ID %d\n", pi.dwProcessId);
    }

    // is this a race condition!!!???
    if (!DeviceIoControl(hDriver,
			    IOCTL_ENABLE_IOPM_ON_PROCESSID,
			    &pi.dwProcessId, sizeof(pi.dwProcessId),
			    NULL, 0, &nBytesRet, NULL)) {

        ShowErr(EGO": DeviceIoControl(IOCTL_ENABLE_IOPM_ON_PROCESSID)");
	fprintf(stderr, EGO": driver "DRIVER_NAME" control error!\n");
        goto abort;
    } else {
	if (bVerbose) fprintf(stderr,
               EGO": driver "DRIVER_NAME" set IOPM for process ID %d\n",
	       pi.dwProcessId);
    }

    // wait until child process exits
    WaitForSingleObject(pi.hProcess, INFINITE);

    if (bVerbose) fprintf(stderr, EGO": child terminated\n");

    // give us the same I/O rights
    pi.dwProcessId = GetCurrentProcessId();
    if (bVerbose) fprintf(stderr, EGO": parent process ID %ld\n", pi.dwProcessId);
    if (!DeviceIoControl(hDriver,
			    IOCTL_ENABLE_IOPM_ON_PROCESSID,
			    &pi.dwProcessId, sizeof(pi.dwProcessId),
			    NULL, 0, &nBytesRet, NULL)) {

        ShowErr(EGO": DeviceIoControl(IOCTL_ENABLE_IOPM_ON_PROCESSID)");
        goto abort;
    }
    CloseHandle(hDriver);

#define _pp_inp(x)    _inp((unsigned short)(x))
#define _pp_outp(x,y) _outp((unsigned short)(x),((int)(y)))
    _try {
        _pp_inp(0x378);  // do "test" I/O read
        fprintf(stderr, "OK 0x378\n");
    } _except( EXCEPTION_EXECUTE_HANDLER ) {
        fprintf(stderr, "failed 0x378!\n");
    }
    _try {
        _pp_inp(0x3bc);  // do "test" I/O read
        fprintf(stderr, "OK 0x3bc\n");
    } _except( EXCEPTION_EXECUTE_HANDLER ) {
        fprintf(stderr, "failed 0x3bc!\n");
    }

    CloseHandle( pi.hProcess );
    CloseHandle( pi.hThread );

    return 0;

abort:
    if (hDriver) CloseHandle(hDriver);
    return 1;
}
